/*
 * include/revirt/rlog.h - definitions relating to the revirt logging system.
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef _RLOG_H
#define _RLOG_H

//#include <linux/sched.h>

#include <revirt/rlog_min.h>
#include <revirt/replay.h>
/*
 * Common definitions, used by both kernel and userland
 */
#define RLOG_SCHEDULE_ENTRY (0x1000)

#define ENTRY_SCHED  (0 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_FORK   (1 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_EXIT   (2 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_TOP    (3 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_SIGNAL	(4 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_EXECPARM (5 | RLOG_SCHEDULE_ENTRY)

//#define ENTRY_PUT_USER	15 // in rlog_min.h
#define ENTRY_COPY_USER	16
#define ENTRY_SYSRET	17
#define ENTRY_CLEAR_USER 18
#define ENTRY_COPY_USER_NODATA 19
#define ENTRY_SIGFRAME 20
#define ENTRY_RT_SIGFRAME 21
#define ENTRY_MARKER 22
#define ENTRY_SENDERIP 23

#define REPLAY_PBUF_MAX_SIZE 10

struct replay_options_xreplay_st {
    int packetrate;
    int packetsize;
};

#ifdef __KERNEL__
typedef  __kernel_fd_set __replay_fd_set;
#else
typedef struct {
	unsigned long fds_bits[32];
} __replay_fd_set;
#endif

struct replay_options_st {
    unsigned int logflags;
    unsigned int replayflags;
    struct replay_options_xreplay_st xreplay;
    int network_fd;
    __replay_fd_set unlogged_fds;
};

#define REPLAY_OPT_LOG_WHOLE_SIGFRAME	(1 << 0)
#define REPLAY_OPT_LOG_WHOLE_PTREGS		(1 << 1)
#define REPLAY_OPT_LOG_WHOLE_SIGINFO	(1 << 2)
#define REPLAY_OPT_LOG_SYSCALL_LOGS_BRANCHCOUNT	(1 << 3)
#define REPLAY_OPT_LOG_ALL_SYSCALLS		(1 << 4)
#define REPLAY_OPT_LOG_LOG_SENDERIP		(1 << 5)
#define REPLAY_OPT_REPLAY_DEBUG		(1 << 8)
#define	REPLAY_OPT_ZTRACE_DEBUG		(1 << 9)
#define REPLAY_OPT_THROTTLE_NETWORK	(1 << 10)
#define REPLAY_OPT_VALUE		(1 << 16)
#define REPLAY_OPT_PACKETRATE		(REPLAY_OPT_VALUE | 1)
#define REPLAY_OPT_PACKETSIZE		(REPLAY_OPT_VALUE | 2)
#define REPLAY_OPT_NETWORKFD		(REPLAY_OPT_VALUE | 3)
#define REPLAY_OPT_UNLOGGEDFDS		(REPLAY_OPT_VALUE | 4)

#define REPLAY_OPT_LOGMASK (0xff)
#define REPLAY_OPT_REPLAYMASK (0xff << 8)

//#define REPLAY_WHOLE_SIGFRAME
//#define REPLAY_WHOLE_PTREGS
//#define REPLAY_WHOLE_SIGINFO
//#define REPLAY_SYSCALL_LOGS_BRANCHCOUNT
//#define REPLAY_ALL_SYSCALLS
//#define REPLAY_LOG_SENDERIP
#define REPLAY_THROTTLE_NETWORK

#define IS_SCHED(x) ( (x) & RLOG_SCHEDULE_ENTRY )

//#define UML_NETWORK_FD 0x44 // 2.2.20 guest
#define UML_NETWORK_FD 0x56 // 2.4.18 guest
#define UML_IS_NETWORK_FD(fd) ((fd)==(replay_options.network_fd))
#define UML_IS_UNLOGGED_FD(fd) (FD_ISSET((fd), &replay_options.unlogged_fds))

#ifndef __KERNEL__
typedef long long s64;
#endif

struct replay_ctrdata {
	s64 hw_branch;
	unsigned int hw_intr;
	unsigned int ex_pf;
	unsigned int ex_dev;
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	int exceptions[20];
	unsigned long softints;
#endif
};

struct replay_senderip {
	unsigned long ip;
};

//#ifdef CONFIG_MPENTIUM4
#if 0
#define ADJ_BC(x) ((x).hw_branch)
#else
/*
 * During replay, for perfctr overflow, we change the branch count; if we
 * want to cross-check branchcount at syscalls then, we need to compensate
 * for that.  Use ADJ_BC_R only if x was copied during replay.
 */
#define ADJ_BC(x) ((x).hw_branch - (x).hw_intr - (x).ex_pf )
#define ADJ_BC_R(x) ((x).hw_branch - (x).hw_intr - (x).ex_pf + zt_perferr)
#endif

struct rlog_header {
	int type;
	unsigned long len;
};


struct rlog_exit_entry {
	short type;
	int pid;
};

struct rlog_schedule_entry {
	short type;
	unsigned int flags;
	int pid;
	unsigned long eip;
	int data;
	struct replay_ctrdata ctr;
	unsigned int edi;
	unsigned int nrxuser;
	unsigned long wchan;
};

struct rlog_entry_fork {
	short type;
	int pid;
	int child;
};

#if 0 // in rlog_min.h
struct rlog_put_user_entry {
	short type;
	unsigned long xx, pptr;
	short ssize, rretval;
	int pid, syscall;
};
#endif

struct rlog_copy_user_entry {
	short type;
	unsigned long tto;
	unsigned long ssize;
	int pid;
	int nrsyscalls;
	int syscall;
};

struct rlog_clear_user_entry {
	short type;
	unsigned long to;
	unsigned long n;
	int pid, syscall;
};

# include <asm/ptrace.h>
//struct pt_regs { int i; };

struct __min_pt_regs { unsigned int eax; };

struct rlog_entry_sysret {
	short type;
	int pid;
	int nrxuser;
	int syscall;
#ifdef REPLAY_WHOLE_PTREGS
	struct pt_regs regs;
#else
	struct __min_pt_regs regs;
#endif
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
	struct replay_ctrdata ctr;
#endif
};


/*
 * Stupid, stupid, stupid...
 */
#ifdef __KERNEL__

#include <linux/types.h>
#include <linux/time.h>
#include <linux/times.h>
#include <asm/signal.h>

#else

#include <linux/posix_types.h>

#define _NSIG_K         64
#define _NSIG_BPW_K     32
#define _NSIG_WORDS_K   (_NSIG_K / _NSIG_BPW_K)

typedef struct {
        unsigned long sig[_NSIG_WORDS_K];
} __kernel_sigset_t;

/* from linux/time.h */
struct __kernel_timeval {
	__kernel_time_t		tv_sec;
	__kernel_suseconds_t	tv_usec;
};

struct __kernel_timezone {
        int     tz_minuteswest; /* minutes west of Greenwich */
        int     tz_dsttime;     /* type of dst correction */
};

struct __kernel_itimerval {
        struct  __kernel_timeval it_interval;    /* timer interval */
        struct  __kernel_timeval it_value;       /* current value */
};

/* from linux/times.h */

struct __kernel_tms {
    __kernel_clock_t tms_utime;
    __kernel_clock_t tms_stime;
    __kernel_clock_t tms_cutime;
    __kernel_clock_t tms_cstime;
};

#endif

struct rlog_marker_entry {
	short type;
	int mark;
#ifdef __KERNEL__
	struct timeval tv;
#else
	struct __kernel_timeval tv;
#endif
};

struct rlog_entry_signal {
	short type;
	int pid;
//	int n;		/* Number of signals in the queue */
#ifndef REPLAY_WHOLE_SIGINFO
	int si_signo, si_errno, si_code;
	int _pad[2];
#endif
#ifdef __KERNEL__	
	sigset_t blocked;
	sigset_t pending_signal;
	sigset_t dfl;
#else
	__kernel_sigset_t blocked;
	__kernel_sigset_t pending_signal;
	__kernel_sigset_t dfl;
#endif
};

#include <asm/sigcontext.h>

#ifdef __KERNEL__

# include <asm/signal.h>
# include <asm/siginfo.h>
# include <asm/ucontext.h>

#else  /* !__KERNEL__ */
# define _NSIG           64
# define _NSIG_BPW       32
# define _NSIG_WORDS     (_NSIG / _NSIG_BPW)

# define SI_MAX_SIZE     128
# define SI_PAD_SIZE     ((SI_MAX_SIZE/sizeof(int)) - 3)

typedef union __kernel_sigval {
	int sival_int;
	void * sival_ptr;
} __kernel_sigval_t;

struct __kernel_siginfo {
	int si_signo;
	int si_errno;
	int si_code;

	union {
		int _pad[SI_PAD_SIZE];
		struct {
			__kernel_pid_t _pid;
			__kernel_uid_t _uid;
		} _kill;

		struct {
			unsigned int _timer1;
			unsigned int _timer2;
		} _timer;

		struct {
			__kernel_pid_t _pid;
			__kernel_uid_t _uid;
			__kernel_sigval_t _sigval;
		} _rt;

		struct {
			__kernel_pid_t _pid;
			__kernel_uid_t _uid;
			int _status;
			__kernel_clock_t _utime;
			__kernel_clock_t _stime;
		} _sigchld;

		struct {
			void * _addr;
		} _sigfault;

		struct {
			int _band;
			int _fd;
		} _sigpoll;
	} _sifields;
};

typedef struct __kernel_sigaltstack {
	void * ss_sp;
	int ss_flags;
	__kernel_size_t ss_size;
} __kernel_stack_t;

struct __kernel_ucontext {
	unsigned long uc_flags;
	struct __kernel_ucontext * uc_link;
	__kernel_stack_t	uc_stack;
	struct sigcontext uc_mcontext;
	__kernel_sigset_t uc_sigmask; // defined above
};

#endif /* __KERNEL__ */


/* from arch/i386/kernel/signal.c */
struct __kernel_sigframe {
	char * pretcode;
	int sig;
	struct sigcontext sc;
	struct _fpstate fpstate;
	unsigned long extramask[_NSIG_WORDS-1];
	char retcode[8];
};

struct __kernel_rt_sigframe {
	char *pretcode;
        int sig;
#ifdef __KERNEL__
        struct siginfo *pinfo;
#else
        struct __kernel_siginfo *pinfo;
#endif
        void *puc;
#ifdef __KERNEL__
        struct siginfo info;
        struct ucontext uc;
#else
        struct __kernel_siginfo info;
        struct __kernel_ucontext uc;
#endif
        struct _fpstate fpstate;
        char retcode[8];
};

/*
 * struct sigrame is declared inside signal.c, so even the kernel needs
 * to re-declare it here...
 */
struct rlog_entry_sigframe {
	short type;
	int pid;
	struct __kernel_sigframe frame;
};

struct rlog_entry_rt_sigframe {
	short type;
	int pid;
#ifdef REPLAY_WHOLE_SIGFRAME
	struct __kernel_rt_sigframe frame;
#else
	int sig;
	unsigned long trapno; // frame.uc.uc_mcontext.trapno
#endif
};

struct rlog_entry_scall_lexecve {
	int type;
	int pid;
	int fname_index;
	int argv_index;
	int env_index;
};

struct rlog_entry_scall_socket {
	int syscall;
	int call;
	int err;
	int len;
};


struct rlog_entry_scall_select {
	int syscall;
	int ret; 
#ifdef __KERNEL__
	fd_set inp, outp, exp;
	struct timeval tv;
#else
	__kernel_fd_set inp, outp, exp;
	struct __kernel_timeval tv;
#endif
};

struct rlog_entry_scall_times {
	int syscall;
	int ret;
#ifdef __KERNEL__
	struct tms tbuf;
#else
	struct __kernel_tms tbuf;
#endif
};

struct rlog_entry_scall_gettimeofday {
	int syscall;
	int err;
#ifdef __KERNEL__
	struct timeval tv;
	struct timezone tz;	/* linux/time.h */
#else
	struct __kernel_timeval tv;
	struct __kernel_timezone tz;	/* linux/time.h */
#endif
};

struct rlog_entry_scall_read {
	int syscall;
	int len;
};

struct rlog_entry_scall_itimerval {
	int syscall;
	int which;
	int ret;
#ifdef __KERNEL__
	struct itimerval itv; /* linux/time.h */
#else
	struct __kernel_itimerval itv; /* linux/time.h */
#endif
};

struct rlog_entry_scall_time {
	int syscall;
	int ret;
};


/*
 * Kernel-only definitions
 */

#ifdef __KERNEL__

#define LOGPAGESIZE 4096
#define NLOGSPERPAGE (LOGPAGESIZE/sizeof(struct log_schedule_entry))

#include <revirt/syscalls.h>

void save_perfctr(struct task_struct *prev);

#include <linux/uio.h>

//int demon_write_log(char * data, int type, int len);
int demon_writev_log(int type, const struct iovec *v, int n);
int demon_read_open(char * fname);
int demon_read_log(struct rlog_header * rh);

void demon_log_fork(struct task_struct *parent,struct task_struct *child);
void record_fork(struct task_struct *parent,struct task_struct *child);
void demon_replay_fork(struct task_struct *parent,struct task_struct *child);

void copy_branchdata(struct task_struct *from, struct replay_ctrdata *to);
void ztrace_yield_to(struct task_struct * tsk);


//extern int replay_exceptions[20];
extern int softints;
extern int replay_pf;
extern int replay_dev;
extern unsigned int replay_interrupts;
extern unsigned int replay_exceptions;
extern struct replay_options_st replay_options;

extern struct rlog_schedule_entry *replay_log;
extern int last_uid;
extern int replay_ready;
extern unsigned int replay_edi;

extern s64 zt_perferr;

extern char replay_scall_buf[];
# endif /* __KERNEL__ */

#endif /* _RLOG_H */
