/*
 * proc_rlog.c - Functions for /proc/rlogd 
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/poll.h>

#include <asm/uaccess.h>
#include <asm/io.h>

/*
 * Needed by kdemon() for procfs
 */
#include <linux/stat.h>
#include <linux/proc_fs.h>


extern wait_queue_head_t revirt_log_read_wait;




extern int do_replaylog(int type, char * buf, int count);

static int proc_rlog_open(struct inode * inode, struct file * file)
{
	return do_replaylog(1,NULL,0);
}

static int proc_rlog_release(struct inode * inode, struct file * file)
{
	(void) do_replaylog(0,NULL,0);
	return 0;
}

static ssize_t proc_rlog_read(struct file * file, char * buf,
			 size_t count, loff_t *ppos)
{
	return do_replaylog(2,buf,count);
}

static unsigned int proc_rlog_poll(struct file *file, poll_table * wait)
{
	int n, ret=0;
	poll_wait(file, &revirt_log_read_wait, wait);
	if ((n=do_replaylog(9, 0, 0))) {
		ret= POLLIN | POLLRDNORM;
	}
	//printk("do_replaylog(POLL) returns %d\n", n);
	return ret;
}

struct file_operations proc_rlog_operations = {
	read:		proc_rlog_read,
	poll:		proc_rlog_poll,
	open:		proc_rlog_open,
	release:	proc_rlog_release,
};

/*
 * We probably want to put this in a module sometime
 */

int replay_ready = 0;

void  proc_rlog_init(void) {
        struct proc_dir_entry *entry;

        /*
         * Create /proc/rlog
         */
        entry=create_proc_entry("rlog", S_IRUSR, &proc_root);
        if(entry)
                entry->proc_fops = &proc_rlog_operations;
	replay_ready = 1;
}
